@@ -1,4 +1,4 @@
-##  R routines for the package mgcv (c) Simon Wood 2000-2008
+##  R routines for the package mgcv (c) Simon Wood 2000-2009
 
 ##  This file is primarily concerned with defining classes of smoother,
 ##  via constructor methods and prediction matrix methods. There are
@@ -1296,11 +1296,30 @@ smooth.construct2 <- function(object,data,knots) {
   object <- smooth.construct(object,dk$data,dk$knots)
   ind <- attr(dk$data,"index") ## repeats index 
   if (!is.null(ind)) { ## unpack the model matrix
+    offs <- attr(object$X,"offset")
     object$X <- object$X[ind,]
+    if (!is.null(offs)) attr(object$X,"offset") <- offs[ind]
   } 
   object
 }
 
+smooth.construct3 <- function(object,data,knots) {
+## This routine does not require that `data' contains only
+## the evaluated `object$term's and the `by' variable... it
+## obtains such a data object from `data' and also deals with
+## multiple evaluations at the same covariate points efficiently
+## In contrast to smooth.constuct2 it returns an object in which
+## `X' contains the rows required to make the full model matrix,
+## and ind[i] tells you which row of `X' is the ith row of the
+## full model matrix. If `ind' is NULL then `X' is the full model matrix. 
+  dk <- ExtractData(object,data,knots) 
+  object <- smooth.construct(object,dk$data,dk$knots)
+  ind <- attr(dk$data,"index") ## repeats index 
+  object$ind <- ind
+  object
+}
+
+
 Predict.matrix <- function(object,data) UseMethod("Predict.matrix")
 
 Predict.matrix2 <- function(object,data) {
@@ -1308,11 +1327,21 @@ Predict.matrix2 <- function(object,data) {
    X <- Predict.matrix(object,dk$data)
    ind <- attr(dk$data,"index") ## repeats index
    if (!is.null(ind)) { ## unpack the model matrix
+     offs <- attr(X,"offset")
      X <- X[ind,]
+     if (!is.null(offs)) attr(X,"offset") <- offs[ind]
    } 
    X
 }
 
+Predict.matrix3 <- function(object,data) {
+## version of Predict.matrix matching smooth.construct3
+   dk <- ExtractData(object,data,NULL) 
+   X <- Predict.matrix(object,dk$data)
+   ind <- attr(dk$data,"index") ## repeats index
+   list(X=X,ind=ind)
+}
+
 
 
 ExtractData <- function(object,data,knots) {
@@ -1368,6 +1397,211 @@ smoothCon <- function(object,data,knots,absorb.cons=FALSE,scale.penalty=TRUE,n=n
 ## in which case a list will do.
 ## If present dataX specifies the data to be used to set up the model matrix, given the 
 ## basis set up using data (but n same for both).
+{ sm <- smooth.construct3(object,data,knots)
+  if (!is.null(attr(sm,"qrc"))) warning("smooth objects should not have a qrc attribute.")
+ 
+  ## automatically produce centering constraint...
+  ## must be done here on original model matrix to ensure same
+  ## basis for all `id' linked terms
+  if (is.null(sm$C)) {
+    sm$C <- matrix(colSums(sm$X),1,ncol(sm$X))
+    conSupplied <- FALSE
+  } else conSupplied <- TRUE
+
+  ## set df fields (pre-constraint)...
+  if (is.null(sm$df)) sm$df <- sm$bs.dim
+
+  ## automatically discard penalties for fixed terms...
+  if (!is.null(object$fixed)&&object$fixed) {
+    sm$S <- NULL
+  }
+
+  ## The following is intended to make scaling `nice' for better gamm performance.
+  ## Note that this takes place before any resetting of the model matrix, and 
+  ## any `by' variable handling. From a `gamm' perspective this is not ideal, 
+  ## but to do otherwise would mess up the meaning of smoothing parameters
+  ## sufficiently that linking terms via `id's would not work properly (they 
+  ## would have the same basis, but different penalties)
+  if (scale.penalty && length(sm$S)>0 && is.null(sm$no.rescale)) # then the penalty coefficient matrix is rescaled
+  {  maXX <- mean(abs(t(sm$X)%*%sm$X)) # `size' of X'X
+      for (i in 1:length(sm$S)) {
+        maS <- mean(abs(sm$S[[i]]))
+        sm$S[[i]] <- sm$S[[i]] * maXX / maS
+      }
+  } 
+
+  ## check whether different data to be used for basis setup
+  ## and model matrix... 
+  if (!is.null(dataX)) { er <- Predict.matrix3(sm,dataX) 
+    sm$X <- er$X
+    sm$ind <- er$ind
+    rm(er)
+  }
+
+  ## check whether smooth called with matrix argument
+  if ((is.null(sm$ind)&&nrow(sm$X)!=n)||(!is.null(sm$ind)&&length(sm$ind)!=n)) { 
+    matrixArg <- TRUE 
+    ## now get the number of columns in the matrix argument...
+    if (is.null(sm$ind)) q <- nrow(sm$X)/n else q <- length(sm$ind)/n
+    if (!is.null(sm$by.done)) warning("handling `by' variables in smooth constructors may not work with the summation convention ")
+  } else {
+    matrixArg <- FALSE
+    if (!is.null(sm$ind)) {  ## unpack model matrix + any offset
+      offs <- attr(sm$X,"offset")
+      sm$X <- sm$X[sm$ind,]      
+      if (!is.null(offs)) attr(sm$X,"offset") <- offs[sm$ind]
+    }
+  }
+  offs <- NULL
+  ## pick up "by variables" now, and handle summation convention ...
+  if (matrixArg||(object$by!="NA"&&is.null(sm$by.done))) 
+  { if (is.null(dataX)) by <- get.var(object$by,data) 
+    else by <- get.var(object$by,dataX)
+    if (matrixArg&&is.null(by)) { ## then by to be taken as sequence of 1s
+      if (is.null(sm$ind)) by <- rep(1,nrow(sm$X)) else by <- rep(1,length(sm$ind))
+    }
+    if (is.null(by)) stop("Can't find by variable")
+    offs <- attr(sm$X,"offset")
+    if (is.factor(by)) { 
+      if (matrixArg) stop("factor `by' variables can not be used with matrix arguments.")
+      sml <- list()
+      lev <- levels(by)
+      for (j in 1:length(lev)) {
+        sml[[j]] <- sm  ## replicate smooth for each factor level
+        by.dum <- as.numeric(lev[j]==by)
+        sml[[j]]$X <- by.dum*sm$X  ## multiply model matrix by dummy for level
+        sml[[j]]$by.level <- lev[j] ## store level
+        sml[[j]]$label <- paste(sm$label,":",object$by,lev[j],sep="") 
+        if (!is.null(offs)) {
+          attr(sml[[j]]$X,"offset") <- offs*by.dum
+        }
+      }
+    } else { ## not a factor by variable
+      sml <- list(sm)
+      if ((is.null(sm$ind)&&length(by)!=nrow(sm$X))||
+          (!is.null(sm$ind)&&length(by)!=length(sm$ind))) stop("`by' variable must be same dimension as smooth arguments")
+     
+      if (matrixArg) { ## arguments are matrices => summation convention used
+        if (is.null(sm$ind)) { ## then the sm$X is in unpacked form
+          sml[[1]]$X <- as.numeric(by)*sm$X ## normal `by' handling
+          ## Now do the summation stuff....
+          ind <- 1:n 
+          X <- sml[[1]]$X[ind,]
+          for (i in 2:q) {
+            ind <- ind + n
+            X <- X + sml[[1]]$X[ind,]
+          }
+          sml[[1]]$X <- X
+          if (!is.null(offs)) { ## deal with any term specific offset (i.e. sum it too)
+            offs <- attr(sm$X,"offset")*as.numeric(by) ## by variable multiplied version
+            ind <- 1:n 
+            offX <- offs[ind,]
+            for (i in 2:q) {
+              ind <- ind + n
+              offX <- offX + offs[ind,]
+            }
+            attr(sml[[1]]$X,"offset") <- offX
+          } ## end of term specific offset handling
+        } else { ## model sm$X is in packed form to save memory
+          ind <- 0:(q-1)*n
+          offs <- attr(sm$X,"offset")
+          if (!is.null(offs)) offX <- rep(0,n) else offX <- NULL 
+          sml[[1]]$X <- matrix(0,n,ncol(sm$X))  ## something wrong with this case!!
+          for (i in 1:n) { ## in this case have to work down the rows
+            ind <- ind + 1
+            sml[[1]]$X[i,] <- colSums(by[ind]*sm$X[sm$ind[ind],]) 
+            if (!is.null(offs)) {
+              offX[i] <- sum(offs[sm$ind[ind]]*by[ind])
+            }      
+          } ## finished all rows
+          attr(sml[[1]]$X,"offset") <- offX
+        } 
+      } else {  ## arguments not matrices => not in packed form + no summation needed 
+        sml[[1]]$X <- as.numeric(by)*sm$X
+        if (!is.null(offs)) attr(sml[[1]]$X,"offset") <- offs*as.numeric(by)
+      }
+
+      sml[[1]]$label <- paste(sm$label,":",object$by,sep="") 
+     
+      ## test for cases where no centring constraint on the smooth is needed. 
+      if (!conSupplied) {
+        if (matrixArg) {
+          ##q <- nrow(sml[[1]]$X)/n
+          L1 <- matrix(by,n,q)%*%rep(1,q)
+          if (sd(L1)>mean(L1)*.Machine$double.eps*1000) sml[[1]]$C <- sm$C <- matrix(0,0,1) 
+          else sml[[1]]$meanL1 <- mean(L1) ## store mean of L1 for use when adding intecept variability
+        } else { ## numeric `by' -- constraint only needed if constant
+          if (sd(by)>mean(by)*.Machine$double.eps*1000) sml[[1]]$C <- sm$C <- matrix(0,0,1)   
+        }
+      } ## end of constraint removal
+    }
+  } else {
+    sml <- list(sm)
+  }
+
+  
+  ## absorb constraints.....
+  if (absorb.cons)
+  { k<-ncol(sm$X)
+    j<-nrow(sm$C)
+    if (j>0) # there are constraints
+    { qrc<-qr(t(sm$C))
+      for (i in 1:length(sml)) { ## loop through smooth list
+        if (length(sm$S)>0)
+        for (l in 1:length(sm$S)) # tensor product terms have > 1 penalty 
+        { ZSZ<-qr.qty(qrc,sm$S[[l]])[(j+1):k,]
+          sml[[i]]$S[[l]]<-t(qr.qty(qrc,t(ZSZ))[(j+1):k,])
+        }
+        sml[[i]]$X<-t(qr.qy(qrc,t(sml[[i]]$X))[(j+1):k,])
+        attr(sml[[i]],"qrc") <- qrc
+        attr(sml[[i]],"nCons") <- j;
+        sml[[i]]$C <- NULL
+        sml[[i]]$rank <- pmin(sm$rank,k-j)
+        sml[[i]]$df <- sml[[i]]$df - j
+        ## ... so qr.qy(attr(sm,"qrc"),c(rep(0,nrow(sm$C)),b)) gives original para.'s
+     } ## end smooth list loop
+   } else { ## no constraints
+     for (i in 1:length(sml)) {
+       attr(sml[[i]],"qrc") <- "no constraints"
+       attr(sml[[i]],"nCons") <- 0;
+     }
+   } ## end else 
+  } else for (i in 1:length(sml)) attr(sml[[i]],"qrc") <-NULL ## no absorption
+
+  ## The idea here is that term selection can be accomplished as part of fitting 
+  ## by applying penalties to the null space of the penalty... 
+
+  if (null.space.penalty) { ## then an extra penalty on the un-penalized space should be added 
+    St <- sml[[1]]$S[[1]]
+    if (length(sml[[1]]$S)>1) for (i in 1:length(sml[[1]]$S)) St <- St + sml[[1]]$S[[i]]
+    es <- eigen(St,symmetric=TRUE)
+    ind <- es$values<max(es$values)*.Machine$double.eps^.66
+    if (sum(ind)) { ## then there is an unpenalized space remaining
+      U <- es$vectors[,ind,drop=FALSE]
+      Sf <- U%*%t(U) ## penalty for the unpenalized components
+      M <- length(sm$S)
+      for (i in 1:length(sml)) {
+        sml[[i]]$S[[M+1]] <- Sf
+        sml[[i]]$rank[M+1] <- sum(ind)
+      }
+    }
+  }
+
+  sml
+} ## end of smoothCon
+
+
+smoothCon1 <- function(object,data,knots,absorb.cons=FALSE,scale.penalty=TRUE,n=nrow(data),
+                      dataX = NULL,null.space.penalty = FALSE)
+## original version .... 
+## wrapper function which calls smooth.construct methods, but can then modify
+## the parameterization used. If absorb.cons==TRUE then a constraint free
+## parameterization is used. 
+## Handles `by' variables, and summation convention.
+## Note that `data' must be a data.frame or model.frame, unless n is provided explicitly, 
+## in which case a list will do.
+## If present dataX specifies the data to be used to set up the model matrix, given the 
+## basis set up using data (but n same for both).
 { sm <- smooth.construct2(object,data,knots)
   if (!is.null(attr(sm,"qrc"))) warning("smooth objects should not have a qrc attribute.")
  
@@ -1467,7 +1701,7 @@ smoothCon <- function(object,data,knots,absorb.cons=FALSE,scale.penalty=TRUE,n=n
       X <- X + sml[[1]]$X[ind,]
     }
     sml[[1]]$X <- X
-    if (!is.null(offs)) { ## deal with any term specific offset
+    if (!is.null(offs)) { ## deal with any term specific offset (i.e. sum it too)
       offs <- attr(sml[[1]]$X,"offset") ## by variable multiplied version
       ind <- 1:n 
       offX <- offs[ind,]
@@ -1529,6 +1763,8 @@ smoothCon <- function(object,data,knots,absorb.cons=FALSE,scale.penalty=TRUE,n=n
   sml
 }
 
+
+
 PredictMat <- function(object,data,n=nrow(data))
 ## wrapper function which calls Predict.matrix and imposes same constraints as 
 ## smoothCon on resulting Prediction Matrix
